class Event < ActiveRecord::Base
  require 'ipaddr'
  set_table_name 'event'
  
  @standard_joins = "LEFT JOIN signature ON event.signature = signature.sig_id
    LEFT JOIN iphdr ON event.sid = iphdr.sid AND event.cid = iphdr.cid
    LEFT JOIN tcphdr ON event.sid = tcphdr.sid AND event.cid = tcphdr.cid"
  
  # CSV templates
  acts_as_csv_exportable :default, [
    {:event_id => :event_cid}, {:event_sid => :event_sid}, {:timestamp => :formatted_timestamp}, {:priority => :sig_priority},
    {:name => :name}, {:source_address => :formatted_source_address}, {:source_port => :tcp_sport},
    {:destination_address => :formatted_destination_address}, {:destination_port => :tcp_dport}]
  
  # Default rows to return when using WillPaginate
  cattr_reader :per_page
  @@per_page = 100
  
  # columns that can be used for sorting indexed by their friendly name
  @sortable_columns = {
    :sensor             => 'sid',
    :timestamp          => 'timestamp',
    :source_ip          => 'ip_src',
    :source_port        => 'tcp_sport',
    :destination_ip     => 'ip_dst',
    :destination_port   => 'tcp_dport',
    :priority           => 'sig_priority',
    :signature          => 'sig_id',
    :name               => 'name'
  }
  @sortable_columns.default = 'timestamp'
  
  def self.threat_indexes params
    where_clause = Condition.block do |c|
      c.and :timestamp, ">=", params[:start_date] unless params[:start_date].blank?
      # This will override the timestamp above if it exists. In other words if both :start_date and :minutes_ago
      # exist :minutes_ago will win.
      c.and :timestamp, ">=", params[:minutes_ago].to_i.minutes.ago unless params[:minutes_ago].blank?
      c.and :timestamp, "<=", params[:end_date] unless params[:end_date].blank?
      c.and :sig_priority, 1
      c.and "event.sid", params[:sensor].to_i unless params[:sensor].blank?
      c.and :ip_src, IPAddr.new(params[:source_ip]).to_i unless params[:source_ip].blank?
      c.and :tcp_sport, params[:source_port].to_i unless params[:source_port].blank?
      c.and :ip_dst, IPAddr.new(params[:destination_ip]).to_i unless params[:destination_ip].blank?
      c.and :tcp_dport, params[:destination_port].to_i unless params[:destination_port].blank?
      c.and :sig_priority, params[:priority].to_i unless params[:priority].blank?
      c.and :sig_id, params[:signature].to_i unless params[:signature].blank?
      # exclude these from the results
      c.and "event.sid", "!=", params[:exclude_sensor].to_i unless params[:exclude_sensor].blank?
      c.and :ip_src, "!=", IPAddr.new(params[:exclude_source_ip]).to_i unless params[:exclude_source_ip].blank?
      c.and :tcp_sport, "!=", params[:exclude_source_port].to_i unless params[:exclude_source_port].blank?
      c.and :ip_dst, "!=", IPAddr.new(params[:exclude_destination_ip]).to_i unless params[:exclude_destination_ip].blank?
      c.and :tcp_dport, "!=", params[:exclude_destination_port].to_i unless params[:exclude_destination_port].blank?
      c.and :sig_id, "!=", params[:exclude_signature].to_i unless params[:exclude_signature].blank?
    end
    Event.find(
      :all,
      :select => "ip_src, max(timestamp) as lastAttempt, COUNT(sig_priority) AS signature_totals,
        count(distinct(ip_dst)) AS distinct_ips, count(distinct(sig_id)) AS distinctSignatures,
        ( (count(distinct(ip_dst)) * (count(distinct(sig_id)) * 5) ) + (COUNT(sig_priority) * .1) )/5 AS threat_index",
      :joins => @standard_joins,
      :conditions => where_clause,
      :group => "ip_src",
      :order => "threat_index DESC, signature_totals DESC, lastAttempt DESC",
      :limit => params[:limit].blank? ? 10 : params[:limit]
    )
  end
  
  def self.get_top_attackers params
    where_clause = Condition.block do |c|
      c.and :timestamp, ">=", params[:start_date] unless params[:start_date].blank?
      # This will override the timestamp above if it exists. In other words if both :start_date and :minutes_ago
      # exist :minutes_ago will win.
      c.and :timestamp, ">=", params[:minutes_ago].to_i.minutes.ago unless params[:minutes_ago].blank?
      c.and :timestamp, "<=", params[:end_date] unless params[:end_date].blank?
      c.and :ip_src, ">", -1
      c.and "event.sid", params[:sensor].to_i unless params[:sensor].blank?
      c.and :ip_src, IPAddr.new(params[:source_ip]).to_i unless params[:source_ip].blank?
      c.and :tcp_sport, params[:source_port].to_i unless params[:source_port].blank?
      c.and :ip_dst, IPAddr.new(params[:destination_ip]).to_i unless params[:destination_ip].blank?
      c.and :tcp_dport, params[:destination_port].to_i unless params[:destination_port].blank?
      c.and :sig_priority, params[:priority].to_i unless params[:priority].blank?
      c.and :sig_id, params[:signature].to_i unless params[:signature].blank?
      # exclude these from the results
      c.and "event.sid", "!=", params[:exclude_sensor].to_i unless params[:exclude_sensor].blank?
      c.and :ip_src, "!=", IPAddr.new(params[:exclude_source_ip]).to_i unless params[:exclude_source_ip].blank?
      c.and :tcp_sport, "!=", params[:exclude_source_port].to_i unless params[:exclude_source_port].blank?
      c.and :ip_dst, "!=", IPAddr.new(params[:exclude_destination_ip]).to_i unless params[:exclude_destination_ip].blank?
      c.and :tcp_dport, "!=", params[:exclude_destination_port].to_i unless params[:exclude_destination_port].blank?
      c.and :sig_priority, "!=", params[:exclude_priority].to_i unless params[:exclude_priority].blank?
      c.and :sig_id, "!=", params[:exclude_signature].to_i unless params[:exclude_signature].blank?
    end
    Event.find(
      :all,
      :select => "DISTINCT iphdr.ip_src as ip_src, COUNT(event.cid) AS count",
      :joins => @standard_joins,
      :conditions => where_clause,
      :group => "ip_src",
      :order => "count DESC",
      :limit => params[:limit].blank? ? 10 : params[:limit]
    )
  end

  def self.get_top_attacks params
    where_clause = Condition.block do |c|
      c.and :timestamp, ">=", params[:start_date] unless params[:start_date].blank?
      # This will override the timestamp above if it exists. In other words if both :start_date and :minutes_ago
      # exist :minutes_ago will win.
      c.and :timestamp, ">=", params[:minutes_ago].to_i.minutes.ago unless params[:minutes_ago].blank?
      c.and :timestamp, "<=", params[:end_date] unless params[:end_date].blank?
      c.and "event.cid", ">", 0
      c.and "event.sid", params[:sensor].to_i unless params[:sensor].blank?
      c.and :ip_src, IPAddr.new(params[:source_ip]).to_i unless params[:source_ip].blank?
      c.and :tcp_sport, params[:source_port].to_i unless params[:source_port].blank?
      c.and :ip_dst, IPAddr.new(params[:destination_ip]).to_i unless params[:destination_ip].blank?
      c.and :tcp_dport, params[:destination_port].to_i unless params[:destination_port].blank?
      c.and :sig_priority, params[:priority].to_i unless params[:priority].blank?
      c.and :sig_id, params[:signature].to_i unless params[:signature].blank?
      # exclude these from the results
      c.and "event.sid", "!=", params[:exclude_sensor].to_i unless params[:exclude_sensor].blank?
      c.and :ip_src, "!=", IPAddr.new(params[:exclude_source_ip]).to_i unless params[:exclude_source_ip].blank?
      c.and :tcp_sport, "!=", params[:exclude_source_port].to_i unless params[:exclude_source_port].blank?
      c.and :ip_dst, "!=", IPAddr.new(params[:exclude_destination_ip]).to_i unless params[:exclude_destination_ip].blank?
      c.and :tcp_dport, "!=", params[:exclude_destination_port].to_i unless params[:exclude_destination_port].blank?
      c.and :sig_priority, "!=", params[:exclude_priority].to_i unless params[:exclude_priority].blank?
      c.and :sig_id, "!=", params[:exclude_signature].to_i unless params[:exclude_signature].blank?
    end
    Event.find(
      :all,
      :select => "sig_name AS name, sig_id, event.sid, event.cid, count(signature) as count",
      :joins => @standard_joins,
      :conditions => where_clause,
      :group => "signature",
      :order => "count DESC",
      :limit => params[:limit].blank? ? 10 : params[:limit]
    )
  end
  
  def self.get_top_high_priority_attacks params
    where_clause = Condition.block do |c|
      c.and :timestamp, ">=", params[:start_date] unless params[:start_date].blank?
      # This will override the timestamp above if it exists. In other words if both :start_date and :minutes_ago
      # exist :minutes_ago will win.
      c.and :timestamp, ">=", params[:minutes_ago].to_i.minutes.ago unless params[:minutes_ago].blank?
      c.and :timestamp, "<=", params[:end_date] unless params[:end_date].blank?
      c.and :sig_priority, 1
      c.and "event.cid", ">", 0
      c.and "event.sid", params[:sensor].to_i unless params[:sensor].blank?
      c.and :ip_src, IPAddr.new(params[:source_ip]).to_i unless params[:source_ip].blank?
      c.and :tcp_sport, params[:source_port].to_i unless params[:source_port].blank?
      c.and :ip_dst, IPAddr.new(params[:destination_ip]).to_i unless params[:destination_ip].blank?
      c.and :tcp_dport, params[:destination_port].to_i unless params[:destination_port].blank?
      c.and :sig_priority, params[:priority].to_i unless params[:priority].blank?
      c.and :sig_id, params[:signature].to_i unless params[:signature].blank?
      # exclude these from the results
      c.and "event.sid", "!=", params[:exclude_sensor].to_i unless params[:exclude_sensor].blank?
      c.and :ip_src, "!=", IPAddr.new(params[:exclude_source_ip]).to_i unless params[:exclude_source_ip].blank?
      c.and :tcp_sport, "!=", params[:exclude_source_port].to_i unless params[:exclude_source_port].blank?
      c.and :ip_dst, "!=", IPAddr.new(params[:exclude_destination_ip]).to_i unless params[:exclude_destination_ip].blank?
      c.and :tcp_dport, "!=", params[:exclude_destination_port].to_i unless params[:exclude_destination_port].blank?
      c.and :sig_id, "!=", params[:exclude_signature].to_i unless params[:exclude_signature].blank?
    end
    Event.find(
      :all,
      :select => "sig_name AS name, sig_id, event.sid, event.cid, count(signature) as count",
      :joins => @standard_joins,
      :conditions => where_clause,
      :group => "signature",
      :order => "count DESC",
      :limit => params[:limit].blank? ? 10 : params[:limit]
    )
  end
  
  def self.get_high_priority_attacks_by_country params
    countries = {}
    
    where_clause = Condition.block do |c|
      c.and :timestamp, ">=", params[:start_date] unless params[:start_date].blank?
      # This will override the timestamp above if it exists. In other words if both :start_date and :minutes_ago
      # exist :minutes_ago will win.
      c.and :timestamp, ">=", params[:minutes_ago].to_i.minutes.ago unless params[:minutes_ago].blank?
      c.and :timestamp, "<=", params[:end_date] unless params[:end_date].blank?
      c.and :sig_priority, 1
      c.and "event.cid", ">", 0
      c.and "event.sid", params[:sensor].to_i unless params[:sensor].blank?
      c.and :ip_src, IPAddr.new(params[:source_ip]).to_i unless params[:source_ip].blank?
      c.and :tcp_sport, params[:source_port].to_i unless params[:source_port].blank?
      c.and :ip_dst, IPAddr.new(params[:destination_ip]).to_i unless params[:destination_ip].blank?
      c.and :tcp_dport, params[:destination_port].to_i unless params[:destination_port].blank?
      c.and :sig_id, params[:signature].to_i unless params[:signature].blank?
      # exclude these from the results
      c.and "event.sid", "!=", params[:exclude_sensor].to_i unless params[:exclude_sensor].blank?
      c.and :ip_src, "!=", IPAddr.new(params[:exclude_source_ip]).to_i unless params[:exclude_source_ip].blank?
      c.and :tcp_sport, "!=", params[:exclude_source_port].to_i unless params[:exclude_source_port].blank?
      c.and :ip_dst, "!=", IPAddr.new(params[:exclude_destination_ip]).to_i unless params[:exclude_destination_ip].blank?
      c.and :tcp_dport, "!=", params[:exclude_destination_port].to_i unless params[:exclude_destination_port].blank?
      c.and :sig_id, "!=", params[:exclude_signature].to_i unless params[:exclude_signature].blank?
    end
    events = Event.find(
      :all,
      :select => "DISTINCT iphdr.ip_src as ip_src, COUNT(event.cid) AS count",
      :joins => @standard_joins,
      :conditions => where_clause,
      :group => "ip_src",
      :order => "count DESC"
    )
    events.each do |e|
      country = e.country_of_source_address
      country_code = country[3].downcase
      event_count = e.count.to_i
      
      countries[country_code] = {:country_name => country[5], :count => 0} unless countries.has_key? country_code
        
      countries[country_code][:count] += event_count
    end
    countries.sort { |a,b| b[1][:count] <=> a[1][:count] }
  end
  
  ##### Commented out to make sure it wasn't being used. inet_ntoa is a MySQL function and shouldn't be used in a query ######
  #def self.get_attack start_time, end_time, id
  #  sql = "SELECT sig_name AS name, sig_id, event.sid, event.cid, inet_ntoa(ip_src) AS ip_src, inet_ntoa(ip_dst) AS ip_dst,
  #    ip_proto, tcp_sport, tcp_dport
  #    FROM event
  #    LEFT JOIN signature ON event.signature=signature.sig_id
  #    LEFT JOIN iphdr ON event.sid = iphdr.sid AND event.cid = iphdr.cid
  #    LEFT JOIN tcphdr ON event.sid = tcphdr.sid AND event.cid = tcphdr.cid
  #    WHERE event.cid > 0 AND sig_priority = 1 AND
  #    sig_id = ? AND event.timestamp BETWEEN ? AND ?"
  #  Event.find_by_sql [sql, id, start_time, end_time]
  #end

  def self.get_alert_severities params
    where_clause = Condition.block do |c|
      c.and :timestamp, ">=", params[:start_date] unless params[:start_date].blank?
      # This will override the timestamp above if it exists. In other words if both :start_date and :minutes_ago
      # exist :minutes_ago will win.
      c.and :timestamp, ">=", params[:minutes_ago].to_i.minutes.ago unless params[:minutes_ago].blank?
      c.and :timestamp, "<=", params[:end_date] unless params[:end_date].blank?
      c.and "event.sid", params[:sensor].to_i unless params[:sensor].blank?
      c.and :ip_src, IPAddr.new(params[:source_ip]).to_i unless params[:source_ip].blank?
      c.and :tcp_sport, params[:source_port].to_i unless params[:source_port].blank?
      c.and :ip_dst, IPAddr.new(params[:destination_ip]).to_i unless params[:destination_ip].blank?
      c.and :tcp_dport, params[:destination_port].to_i unless params[:destination_port].blank?
      c.and :sig_priority, params[:priority].to_i unless params[:priority].blank?
      c.and :sig_id, params[:signature].to_i unless params[:signature].blank?
      # exclude these from the results
      c.and "event.sid", "!=", params[:exclude_sensor].to_i unless params[:exclude_sensor].blank?
      c.and :ip_src, "!=", IPAddr.new(params[:exclude_source_ip]).to_i unless params[:exclude_source_ip].blank?
      c.and :tcp_sport, "!=", params[:exclude_source_port].to_i unless params[:exclude_source_port].blank?
      c.and :ip_dst, "!=", IPAddr.new(params[:exclude_destination_ip]).to_i unless params[:exclude_destination_ip].blank?
      c.and :tcp_dport, "!=", params[:exclude_destination_port].to_i unless params[:exclude_destination_port].blank?
      c.and :sig_priority, "!=", params[:exclude_priority].to_i unless params[:exclude_priority].blank?
      c.and :sig_id, "!=", params[:exclude_signature].to_i unless params[:exclude_signature].blank?
    end
    Event.find(
      :all,
      :select => "sig_priority, count(sig_priority) as total",
      :joins => @standard_joins,
      :conditions => where_clause,
      :group => "sig_priority"
    )
  end
  
  def self.get_filtered_events params
    # Build the where clause
    where_clause = Condition.block do |c|
      c.and :timestamp, ">=", params[:start_date] unless params[:start_date].blank?
      # This will override the timestamp above if it exists. In other words if both :start_date and :minutes_ago
      # exist :minutes_ago will win.
      c.and :timestamp, ">=", params[:minutes_ago].to_i.minutes.ago unless params[:minutes_ago].blank?
      c.and :timestamp, "<=", params[:end_date] unless params[:end_date].blank?
      c.and "event.sid", params[:sensor].to_i unless params[:sensor].blank?
      c.and :ip_src, IPAddr.new(params[:source_ip]).to_i unless params[:source_ip].blank?
      c.and :tcp_sport, params[:source_port].to_i unless params[:source_port].blank?
      c.and :ip_dst, IPAddr.new(params[:destination_ip]).to_i unless params[:destination_ip].blank?
      c.and :tcp_dport, params[:destination_port].to_i unless params[:destination_port].blank?
      c.and :sig_priority, params[:priority].to_i unless params[:priority].blank?
      c.and :sig_id, params[:signature].to_i unless params[:signature].blank?
      # exclude these from the results
      c.and "event.sid", "!=", params[:exclude_sensor].to_i unless params[:exclude_sensor].blank?
      c.and :ip_src, "!=", IPAddr.new(params[:exclude_source_ip]).to_i unless params[:exclude_source_ip].blank?
      c.and :tcp_sport, "!=", params[:exclude_source_port].to_i unless params[:exclude_source_port].blank?
      c.and :ip_dst, "!=", IPAddr.new(params[:exclude_destination_ip]).to_i unless params[:exclude_destination_ip].blank?
      c.and :tcp_dport, "!=", params[:exclude_destination_port].to_i unless params[:exclude_destination_port].blank?
      c.and :sig_priority, "!=", params[:exclude_priority].to_i unless params[:exclude_priority].blank?
      c.and :sig_id, "!=", params[:exclude_signature].to_i unless params[:exclude_signature].blank?
    end
    Event.paginate(
      :all,
      :select => "sig_priority, sig_name AS name,
        signature.sig_id, event.sid as event_sid, event.cid as event_cid, ip_src, ip_dst,
        ip_proto, tcp_sport, tcp_dport, timestamp",
      :joins => @standard_joins,
      :conditions => where_clause,
      :order => "#{@sortable_columns[params[:order_by].blank? ? '' : params[:order_by].to_sym]}
        #{ 'DESC' unless ( !params[:sort_order].blank? && params[:sort_order].upcase.eql?( 'ASC' ) )}",
      :page => params[:page].blank? ? 1 : params[:page],
      :per_page => params[:limit].blank? ? 100 : params[:limit]
    )
  end
  
  def self.get_references sensor_id, event_id
    Event.find(
      :all,
      :select => "event.cid, event.sid, ref_tag, ref_system_name, sig_name",
      :joins => "LEFT JOIN signature ON event.signature = signature.sig_id
        LEFT JOIN sig_reference ON signature.sig_id = sig_reference.sig_id
        LEFT JOIN reference ON sig_reference.ref_id = reference.ref_id
        LEFT JOIN reference_system ON reference.ref_system_id = reference_system.ref_system_id",
      :conditions => ["event.cid = ? AND event.sid = ?", event_id, sensor_id]
    )
  end
  
  def formatted_source_address
    IPAddr.new(self.ip_src.to_i, Socket::AF_INET).to_s
  end
  
  def formatted_destination_address
    IPAddr.new(self.ip_dst.to_i, Socket::AF_INET).to_s
  end
  
  def formatted_timestamp
    self.timestamp.strftime '%m/%d/%Y %I:%M:%S'
  end
  
  def country_of_source_address
    GeoIP.new('db/GeoIP.dat').country(formatted_source_address)
  end
  
  def country_of_destination_address
    GeoIP.new('db/GeoIP.dat').country(formatted_destination_address)
  end
end
